/* 
  mgcld_indirect.h
  
  This file is part of Monochrome Graphic LCD library (MGLCD) for Arduino
  Uno/Mega 2560.
  
  MGLCD developed by Synapse from 2012 to 2013.  This Library is originally 
  distributed at Team Schaft's Homepage. <http://www3.big.or.jp/~schaft>

  MGLCD is now under beta testing, so specification may be changed 
  in the future.

  MGLCD is free software: you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as published by
  the Free Software Foundation, either version 2.1 of the License, or
  (at your option) any later version.

  MGLCD is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with MGLCD.  If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef MGLCD_INDIRECT_H
#define MGLCD_INDIRECT_H

#include <arduino.h>
#include <Print.h>

#if MGLCD_VERSION != 22
#error versions of this file and header file are different.
#endif

#define MGLCD_COMMUNICATION_ERROR (-128)

// Constats for Backlight function
#define MGLCD_BACKLIGHT_OFF         0
#define MGLCD_BACKLIGHT_ON          1
#define MGLCD_GET_BACKLIGHT_STATE 127

// Constants for Contrast function
#define MGLCD_MAX_CONTRAST          15
#define MGLCD_GET_CURRENT_CONTRAST 126
#define MGLCD_GET_DEFAULT_CONTRAST 127
#define MGLCD_NOT_SET_DEFAULT        0
#define MGLCD_SET_DEFAULT            1

// Constants for ReadButton function
#define MGLCD_BUTTON_OFF       0
#define MGLCD_BUTTON_ON        1
#define MGLCD_GET_BUTTON_NUM 127

// Constants for Led function
#define MGLCD_LED_OFF         0
#define MGLCD_LED_ON          1
#define MGLCD_LED_BLINK       2
#define MGLCD_LED_TOGGLE      3
#define MGLCD_GET_LED_NUM   126
#define MGLCD_GET_LED_STATE 127

class SoftwareSerial;

template<class T> inline void FlushWriteBuffer(T *interface) {}
template<> inline void FlushWriteBuffer(HardwareSerial *interface){ interface->flush(); }

template<class T> class mglcd_indirect : public Print
{
public:
  static const int MaxCommandSize = 16;
  static const int MaxCommandResultSize = 10;
  static const int PreferredBaudRateNum = 14;
  static const unsigned long DefaultBaudRate = 1200;
  
  enum PortCommand {
    SCPing,
    SCChangeBaudRate,
    SCReset,
    SCClearScreen,
    SCSetTextWindow,
    SCSetInvertMode,
    SCSetScrollMode,
    SCLocate,
    SCGetX,
    SCGetY,
    SCClearRestOfLine,
    SCUserChar,
    SCSetPixel,
    SCGetPixel,
    SCLine,
    SCRect,
    SCFillRect,
    SCRoundRect,
    SCCircle,
    SCFillCircle,
    SCBacklight,
    SCContrast,
    SCStartDrawBitmap,
    SCSendBitmap,
    SCReadButton,
    SCLed,
    SCMax
  };
  
  static PROGMEM prog_uchar ParamNum[SCMax][2];
  static PROGMEM prog_uint32_t PreferredBaudRate[PreferredBaudRateNum];

private:
  // This class's original member variables
  static const int XON  = 0x11; // XON character
  static const int XOFF = 0x13; // XOFF character
  static const int STX  = 0x02; // STX character
  static const int ETX  = 0x03; // ETX character
  static const int ACK  = 0x06; // ACK character
  static const int NAK  = 0x15; // NAK character
  static const int ERR  = 0x07; // character which means errors occured in command

  unsigned char command[MaxCommandSize];
  unsigned char CommandResult[MaxCommandResultSize];
  unsigned BitBuffer;
  unsigned char ValidBitNum;
  unsigned long BaudRate;
  unsigned char PingCnt;
  
  // Member variables compatible with the other classes
  unsigned char WindowX1,WindowY1,WindowX2,WindowY2; 
  unsigned char InvertMode;
  unsigned char ScrollMode;
  unsigned char Width;     // Screen width
  unsigned char Height;    // Screen height
  unsigned char HalfWidth; // Half of Width
  unsigned char PageNum;   // Number of pages (character lines)
  unsigned char ColumnNum; // Number of character columns

  // This class's original member functions
  T *pInterface;
  void SendBits(unsigned char data, unsigned char BitNum);
  int ReceiveBits(unsigned char BitNum, unsigned long StartMillis, unsigned time_out);
  signed char SendCommand(unsigned time_out=1000);

public:
  // constructor
  explicit mglcd_indirect(T *p, unsigned long baud=DefaultBaudRate);
  
  // destructor
  ~mglcd_indirect(void) {}
  
  // This class's original member functions
  signed char ChangeBaudRate(unsigned long NewBaudRate);
  signed char Backlight(unsigned char OnOff);
  signed char Contrast(unsigned char val, unsigned char option=MGLCD_SET_DEFAULT);
  signed char ReadButton(unsigned char ButtonNo);
  signed char Led(unsigned char LedNo, unsigned char state=MGLCD_LED_OFF);
  
  // The same member functions compatible with the other classes
  int GetConstructionError(void) { return 0; }
  signed char Reset(unsigned char invert=MGLCD_NON_INVERT, unsigned char scroll=MGLCD_SCROLL);
  signed char ClearScreen(void);
  unsigned char GetWidth(void) { return Width; }
  unsigned char GetHeight(void) { return Height; }
  unsigned char GetPageNum(void) { return PageNum; }
  unsigned char GetLineNum(void) { return PageNum; }
  unsigned char GetColumnNum(void) { return ColumnNum; }
  signed char SetTextWindow(unsigned char x1, unsigned char y1, unsigned char x2, unsigned char y2);
  void GetTextWindow(unsigned char &x1, unsigned char &y1, unsigned char &x2, unsigned char &y2) { x1=WindowX1; y1=WindowY1; x2=WindowX2; y2=WindowY2; }
  signed char SetInvertMode(unsigned char mode);
  unsigned char GetInvertMode(void) { return InvertMode; }
  signed char SetScrollMode(unsigned char mode);
  unsigned char GetScrollMode(void) { return ScrollMode; }
  signed char Locate(unsigned char x, unsigned char y);
  int GetX(void);
  int GetY(void);  
  signed char ClearRestOfLine(void);
  signed char PutChar(unsigned char ch);
  signed char UserChars(unsigned char uc[][5], unsigned char num, unsigned char mem=MGLCD_ROM);
  signed char SetPixel(unsigned char x, unsigned char y, unsigned char color=1);
  signed char GetPixel(unsigned char x, unsigned char y);
  signed char Line(unsigned char x1, unsigned char y1, unsigned char x2, unsigned char y2, unsigned char color=1);
  signed char Rect(unsigned char x1, unsigned char y1, unsigned char x2, unsigned char y2, unsigned char color=1);
  signed char FillRect(unsigned char x1, unsigned char y1, unsigned char x2, unsigned char y2, unsigned char color=1);
  signed char RoundRect(unsigned char x1, unsigned char y1, unsigned char x2, unsigned char y2, unsigned char r, unsigned char color=1);  
  signed char Circle(unsigned char x0, unsigned char y0, unsigned char r, unsigned char color=1);
  signed char FillCircle(unsigned char x0, unsigned char y0, unsigned char r, unsigned char color=1);
  signed char DrawBitmap(unsigned char bmp[], unsigned char x, unsigned char y, unsigned char mem=MGLCD_ROM);
  size_t write(uint8_t); // inherit Print
}; // mglcd_indirect

typedef mglcd_indirect<HardwareSerial> mglcd_serial;
typedef mglcd_indirect<SoftwareSerial> mglcd_SoftwareSerial;

template<class T> PROGMEM prog_uchar mglcd_indirect<T>::ParamNum[SCMax][2] = {
// Left number is number of command parameter.  Right number is number of return parameter.
  { 0, 0 }, // SCPing
  { 4, 0 }, // SCChangeBaudRate
  { 3, 4 }, // SCReset
  { 0, 0 }, // SCClearScreen
  { 4, 0 }, // SCSetTextWindow
  { 1, 0 }, // SCSetInvertMode
  { 1, 0 }, // SCSetScrollMode
  { 2, 0 }, // SCLocate
  { 0, 1 }, // SCGetX
  { 0, 1 }, // SCGetY 
  { 0, 0 }, // SCClearRestOfLine,
  { 6, 0 }, // SCUserChar,
  { 3, 0 }, // SCSetPixel,
  { 2, 1 }, // SCGetPixel,
  { 5, 0 }, // SCLine
  { 5, 0 }, // SCRect
  { 5, 0 }, // SCFillRect
  { 6, 0 }, // SCRoundRect
  { 4, 0 }, // SCCircle
  { 4, 0 }, // SCFillCircle
  { 1, 1 }, // SCBacklight
  { 2, 1 }, // SCContrast
  { 4, 0 }, // SCStartDrawBitmap
  { MaxCommandSize-1,0 }, // SCSendBitmap
  { 1, 1 }, // SCReadButton
  { 2, 1 }, // SCLed
};

template<class T> PROGMEM prog_uint32_t mglcd_indirect<T>::PreferredBaudRate[PreferredBaudRateNum] = {
  300    ,1200   ,2400   ,4800   ,9600   ,
  14400  ,19200  ,28800  ,38400  ,57600  ,
  115200 ,250000 ,500000 ,1000000 
};

template<class T> void mglcd_indirect<T>::SendBits(unsigned char data, unsigned char BitNum)
{
  BitBuffer=(BitBuffer<<BitNum)|data;
  ValidBitNum+=BitNum;
  while(ValidBitNum>=7) {
    pInterface->write(0x80|((BitBuffer>>(ValidBitNum-=7))&0x7f));
  } // while
} // mglcd_indirect<T>::SendBits

template<class T> int mglcd_indirect<T>::ReceiveBits(unsigned char BitNum, unsigned long StartMillis, unsigned time_out)
{
  unsigned char data;
  
  while(ValidBitNum<BitNum) {
    while(pInterface->available()==0) if(millis()-StartMillis>time_out) return MGLCD_COMMUNICATION_ERROR;
    if((data=pInterface->read())&0x80==0) return MGLCD_COMMUNICATION_ERROR;    
    BitBuffer=(BitBuffer<<7)|(data&0x7f);
    ValidBitNum+=7;
  } // while
  ValidBitNum-=BitNum;
  return (BitBuffer>>ValidBitNum)&((1<<BitNum)-1);
} // mglcd_indirect<T>::ReceiveBits

template<class T> signed char mglcd_indirect<T>::SendCommand(unsigned time_out)
{
  unsigned char parity;
  int i;
  unsigned char CommandSize;
  unsigned long StartMillis;
  int data;
  unsigned char CommandResultSize;
  
  if(command[0]>=SCMax) return -1; // Undefined command
  
  // Get and check command size
  CommandSize=(unsigned char)pgm_read_byte_near(&ParamNum[command[0]][0])+1; 
  if(CommandSize>MaxCommandSize) return -2;
    
  while(pInterface->available()) pInterface->read(); // Delete read buffer
  
  pInterface->write(STX);
  ValidBitNum=0;
  SendBits(command[0],6);
  parity=(command[0]>>4)^(command[0]&0xf);

  for(i=1; i<CommandSize; i++) {
    unsigned char b=command[i];
    SendBits(b,8);
    parity^=(b>>4)^(b&0xf);
  } // for i
  
  SendBits(parity,4);
  SendBits(0,6); // Flush BitBuffer
  pInterface->write(ETX);
  FlushWriteBuffer(pInterface);
  
  // Wait for first byte of command result
  StartMillis=millis();
  while(true) {
    data=pInterface->read();
    if(data==ACK) {
      return 0;
    } else if(data==NAK) {
      return MGLCD_COMMUNICATION_ERROR;
    } else if(data==ERR) {
      return -3;
    } else if(data==STX) {
      break;
    } else if(data==ETX) {
      return MGLCD_COMMUNICATION_ERROR;
    } // if
    if(millis()-StartMillis>time_out) return MGLCD_COMMUNICATION_ERROR;
  } // while
  
  // Get command result size
  CommandResultSize=(unsigned char)pgm_read_byte_near(&ParamNum[command[0]][1]);

  // Receive command result
  ValidBitNum=0;
  parity=0;
  for(i=0; i<CommandResultSize; i++) {
    if((data=ReceiveBits(8,StartMillis,time_out))<0) return MGLCD_COMMUNICATION_ERROR;
    CommandResult[i]=data;
    parity^=(data>>4)^(data&0xf);
  } // for i

  // Recieve parity
  if((data=ReceiveBits(4,StartMillis,time_out))<0 || data!=parity) return MGLCD_COMMUNICATION_ERROR;
  
  // Recieve ETX
  while(pInterface->available()==0) if(millis()-StartMillis>time_out) return MGLCD_COMMUNICATION_ERROR;
  if(pInterface->read()!=ETX) return MGLCD_COMMUNICATION_ERROR;
  
  return 0;
} // mglcd_indirect<T>::SendCommand

template<class T> mglcd_indirect<T>::mglcd_indirect(T *p, unsigned long baud) : pInterface(p), PingCnt(0), BaudRate(baud)
{
  p->begin(baud);
} // mglcd_indirect<T>::mglcd_indirect

template<class T> signed char mglcd_indirect<T>::ChangeBaudRate(unsigned long NewBaudRate)
{
  unsigned char result;
  int i;
  
  for(i=0; i<PreferredBaudRateNum+1; i++) {
    if(i>0) pInterface->begin(pgm_read_dword_near(PreferredBaudRate+(i-1)));
    command[0]=SCChangeBaudRate; 
    command[1]=(unsigned char)( NewBaudRate     &0xff);
    command[2]=(unsigned char)((NewBaudRate>> 8)&0xff);
    command[3]=(unsigned char)((NewBaudRate>>16)&0xff);
    command[4]=(unsigned char)((NewBaudRate>>24)&0xff);
    if(SendCommand(100)==0) {
      delay(36);
      pInterface->begin(NewBaudRate);
      BaudRate=NewBaudRate;
      return 0;
    } // if
  } // for i
  
  return MGLCD_COMMUNICATION_ERROR;
} // mglcd_indirect<T>::ChangeBaudRate

template<class T> signed char mglcd_indirect<T>::Backlight(unsigned char OnOff)
{
  signed char result;
  
  command[0]=SCBacklight;
  command[1]=OnOff;
  result=SendCommand();
  if(result!=0) return result;
  return CommandResult[0];
} // mglcd_indirect<T>::Backlight

template<class T> signed char mglcd_indirect<T>::Contrast(unsigned char val, unsigned char option)
{
  signed char result;
  
  command[0]=SCContrast;
  command[1]=val;
  command[2]=option;
  result=SendCommand();
  if(result!=0) return result;
  return CommandResult[0];
} // mglcd_indirect<T>::Contrast

template<class T> signed char mglcd_indirect<T>::Reset(unsigned char invert, unsigned char scroll)
{
  signed char result;
  
  // Initialize port
  FlushWriteBuffer(pInterface);
  delay(36);

  // Change baud rate
  result=ChangeBaudRate(BaudRate);
  if(result!=0) return result;
  
  // Send reset comand
  command[0]=SCReset;
  command[1]=invert;
  command[2]=scroll;
  result=SendCommand();
  if(result!=0) return result;

  // Check MGLCD version  
  if(CommandResult[0]+(int(CommandResult[1])<<8)!=MGLCD_VERSION) return -10;
  
  // Initialize variables
  Width =CommandResult[2];
  Height=CommandResult[3];
  HalfWidth=Width/2;
  PageNum=Height/8;
  ColumnNum=Width/MGLCD_FONT_WIDTH;
  InvertMode=invert;
  ScrollMode=scroll;
  
  return 0;
} // mglcd_indirect<T>::Reset

template<class T> signed char mglcd_indirect<T>::ClearScreen(void)
{
  command[0]=SCClearScreen;
  return SendCommand();
} // mglcd_indirect<T>::ClearScreen

template<class T> signed char mglcd_indirect<T>::SetTextWindow(unsigned char x1, unsigned char y1, unsigned char x2, unsigned char y2)
{
  signed char result;
  
  command[0]=SCSetTextWindow;
  command[1]=x1;
  command[2]=y1;
  command[3]=x2;
  command[4]=y2;
  result=SendCommand();
  if(result==0) {
    WindowX1=x1; WindowY1=y1;
    WindowX2=x2; WindowY2=y2;
  } // if
  
  return result;
} // mglcd_indirect<T>::SetTextWindow

template<class T> signed char mglcd_indirect<T>::SetInvertMode(unsigned char mode)
{
  signed char result;

  command[0]=SCSetInvertMode;
  command[1]=mode;
  result=SendCommand();
  if(result==0) InvertMode=mode;
  return result;
} // mglcd_indirect<T>::SetInvertMode

template<class T> signed char mglcd_indirect<T>::SetScrollMode(unsigned char mode)
{
  signed char result;

  command[0]=SCSetScrollMode;
  command[1]=mode;
  result=SendCommand();
  if(result==0) ScrollMode=mode;
  return result;
} // mglcd_indirect<T>::SetScrollMode

template<class T> signed char mglcd_indirect<T>::Locate(unsigned char x, unsigned char y)
{
  command[0]=SCLocate;
  command[1]=x;
  command[2]=y;
  return SendCommand();
} // mglcd_indirect<T>::Locate

template<class T> int mglcd_indirect<T>::GetX(void)
{
  signed char result;
  
  command[0]=SCGetX;
  result=SendCommand();
  if(result!=0) return result;
  return CommandResult[0];
} // mglcd_indirect<T>::GetX

template<class T> int mglcd_indirect<T>::GetY(void)
{
  signed char result;
  
  command[0]=SCGetY;
  result=SendCommand();
  if(result!=0) return result;

  return CommandResult[0];
} // mglcd_indirect<T>::GetY

template<class T> signed char mglcd_indirect<T>::ClearRestOfLine(void)
{
  command[0]=SCClearRestOfLine;
  return SendCommand();
} // mglcd_indirect<T>::ClearRestOfLine

template<class T> signed char mglcd_indirect<T>::PutChar(unsigned char ch)
{
  signed char result;

  // flush Rx buffer of LCD controller
  if(++PingCnt>=16) {
    PingCnt=0;
    command[0]=SCPing;
    result=SendCommand();
    if(result!=0) return result; 
  } // if
  
  if(ch==STX || ch==ETX || ch==ACK || ch==NAK || ch==XON || ch==XOFF) ch=' ';
  pInterface->write(ch);
  
  return 0;
} // mglcd_indirect<T>::PutChar

template<class T> signed char mglcd_indirect<T>::UserChars(unsigned char uc[][5], unsigned char num, unsigned char mem)
{
  unsigned char ChNo, x;
  signed char result;
  
  if(uc==NULL || num==0 || num>32 || (mem!=MGLCD_ROM && mem!=MGLCD_RAM)) return -10;
  for(ChNo=0; ChNo<32; ChNo++) {
    command[0]=SCUserChar;
    command[1]=ChNo;
    if(ChNo<num) {
      for(x=0; x<5; x++) {
        command[x+2]=mem==MGLCD_ROM ? pgm_read_byte_near(&uc[ChNo][x]) : uc[ChNo][x];
      } // for x
    } else {
      command[2]=0;
      command[3]=0;
      command[4]=0;
      command[5]=0;
      command[6]=0;
    } // if
    result=SendCommand();
    
    if(result!=0) return result;
  } // for ChNo
  
  return 0;
} // mglcd_indirect<T>::UserChars

template<class T> signed char mglcd_indirect<T>::SetPixel(unsigned char x, unsigned char y, unsigned char color)
{
  command[0]=SCSetPixel;
  command[1]=x;
  command[2]=y;
  command[3]=color;
  return SendCommand();
} // mglcd_indirect<T>::SetPixel

template<class T> signed char mglcd_indirect<T>::GetPixel(unsigned char x, unsigned char y)
{
  signed char result;
  
  command[0]=SCGetPixel;
  command[1]=x;
  command[2]=y;
  result=SendCommand();
  if(result!=0) return result;

  return (signed char) CommandResult[0];
} // mglcd_indirect<T>::GetPixel

template<class T> signed char mglcd_indirect<T>::Line(unsigned char x1, unsigned char y1, unsigned char x2, unsigned char y2, unsigned char color)
{
  command[0]=SCLine;
  command[1]=x1;
  command[2]=y1;
  command[3]=x2;
  command[4]=y2;
  command[5]=color;
  return SendCommand();
} // mglcd_indirect<T>::Line

template<class T> signed char mglcd_indirect<T>::Rect(unsigned char x1, unsigned char y1, unsigned char x2, unsigned char y2, unsigned char color)
{

  command[0]=SCRect;
  command[1]=x1;
  command[2]=y1;
  command[3]=x2;
  command[4]=y2;
  command[5]=color;
  return SendCommand();
} // mglcd_indirect<T>::Rect

template<class T> signed char mglcd_indirect<T>::FillRect(unsigned char x1, unsigned char y1, unsigned char x2, unsigned char y2, unsigned char color)
{
  command[0]=SCFillRect;
  command[1]=x1;
  command[2]=y1;
  command[3]=x2;
  command[4]=y2;
  command[5]=color;
  return SendCommand();
} // mglcd_indirect<T>::FillRect

template<class T> signed char mglcd_indirect<T>::RoundRect(unsigned char x1, unsigned char y1, unsigned char x2, unsigned char y2, unsigned char r, unsigned char color)
{

  command[0]=SCRoundRect;
  command[1]=x1;
  command[2]=y1;
  command[3]=x2;
  command[4]=y2;
  command[5]=r;
  command[6]=color;
  return SendCommand();
} // mglcd_indirect<T>::RoundRect

template<class T> signed char mglcd_indirect<T>::Circle(unsigned char x0, unsigned char y0, unsigned char r, unsigned char color)
{
  command[0]=SCCircle;
  command[1]=x0;
  command[2]=y0;
  command[3]=r;
  command[4]=color;
  return SendCommand();
} // mglcd_indirect<T>::Circle

template<class T> signed char mglcd_indirect<T>::FillCircle(unsigned char x0, unsigned char y0, unsigned char r, unsigned char color)
{
  command[0]=SCFillCircle;
  command[1]=x0;
  command[2]=y0;
  command[3]=r;
  command[4]=color;
  return SendCommand();
} // mglcd_indirect<T>::FillCircle

template<class T> signed char mglcd_indirect<T>::DrawBitmap(unsigned char bmp[], unsigned char x, unsigned char y, unsigned char mem)
{
  unsigned char BmpWidth,BmpHeight;
  unsigned DataNum;
  unsigned DataIndex;
  unsigned char i;

  if(mem!=MGLCD_ROM && mem!=MGLCD_RAM) return -10;
  if(mem==MGLCD_ROM) {
    BmpWidth =pgm_read_byte_near(bmp+0);
    BmpHeight=pgm_read_byte_near(bmp+1);
  } else { // mem==MGLCD_RAM
    BmpWidth =bmp[0];
    BmpHeight=bmp[1];
  } // if
  if(x+BmpWidth>Width || y+BmpHeight>Height) return -11;

  command[0]=SCStartDrawBitmap;
  command[1]=BmpWidth;
  command[2]=BmpHeight;
  command[3]=x;
  command[4]=y;
  if(SendCommand()!=0) return MGLCD_COMMUNICATION_ERROR;

  DataNum=BmpWidth*((BmpHeight+7)/8)+2;
  for(DataIndex=2; DataIndex<DataNum; DataIndex+=MaxCommandSize-1) {
    command[0]=SCSendBitmap;
    for(i=0; i<MaxCommandSize-1; i++) {
      command[i+1]=DataIndex+i<DataNum ? (mem==MGLCD_ROM ? pgm_read_byte_near(&bmp[DataIndex+i]) : bmp[DataIndex+i]) : 0;
    } // for i
    if(SendCommand()!=0) return MGLCD_COMMUNICATION_ERROR;
  } // for DataIndex
  
  return 0;
} // mglcd_indirect<T>::DrawBitmap

template<class T> signed char mglcd_indirect<T>::ReadButton(unsigned char ButtonNo)
{
  signed char result;
  
  command[0]=SCReadButton;
  command[1]=ButtonNo;
  result=SendCommand();
  if(result!=0) return result;
  return CommandResult[0];
} // mglcd_indirect<T>::ReadButton

template<class T> signed char mglcd_indirect<T>::Led(unsigned char LedNo, unsigned char state)
{
  unsigned char result;
  
  command[0]=SCLed;
  command[1]=LedNo;
  command[2]=state;
  result=SendCommand();
  if(result!=0) return result;
  return CommandResult[0];
} // mglcd_indirect<T>::Led

template<class T> size_t mglcd_indirect<T>::write(uint8_t ch)
{
  PutChar(ch);
  return 1;
} // mglcd_indirect<T>::write

#endif // #ifdef MGLCD_INDIRECT_H
